__title__ = "accession"
__version__ = "3.0.1"
__description__ = "Tool to submit genomics pipeline outputs to the ENCODE Portal"
__url__ = "https://github.com/ENCODE-DCC/accession"
__uri__ = __url__
__author__ = "Paul Sud"
__email__ = "encode-help@lists.stanford.edu"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2019 ENCODE DCC"
